from logging import Logger

import time
import os
from typing import List

from sidecar.const import Const
from sidecar.file_system import FileSystemService
from sidecar.health_check.app_health_check_configuration import AppHealthCheckConfiguration
from sidecar.utils import CallsLogger


class HealthCheckPreparer:
    def __init__(self,
                 file_system: FileSystemService,
                 logger: Logger,
                 health_check_configuration: AppHealthCheckConfiguration):
        self.health_check_configuration = health_check_configuration
        self._file_system = file_system
        self._logger = logger

    def prepare(self, app_name: str, address: str):
        # From the section of 'healthcheck' in the app yaml
        health_check_configuration = self.health_check_configuration.get_configuration(app_name)

        # If we defined 'wait_for_ports' in the 'healthcheck' section in the app yaml
        target_ports = health_check_configuration.target_ports
        if target_ports:
            app_dir = Const.get_app_folder(app_name=app_name)
            self._file_system.create_folder(app_dir)
            return self._create_default_health_check_script(default_ports=target_ports,
                                                            address=address,
                                                            app_name=app_name)

        # If we defined 'script' in the 'healthcheck' section in the app yaml
        if health_check_configuration.script_name:
            script_file_path = Const.get_health_check_file(app_name=app_name,
                                                           script_name=health_check_configuration.script_name)

            return ['{script_file_path} {ip_address}'.format(
                script_file_path=script_file_path,
                ip_address=address)]

        return None

    @CallsLogger.wrap
    def _create_default_health_check_script(self,
                                            default_ports: List[str],
                                            app_name: str,
                                            address: str):
        script_file_path = Const.get_health_check_file(app_name=app_name,
                                                       script_name="default-{0}-hc-{1}.sh".
                                                       format(app_name, str(time.time())))
        lines = list()

        lines.append('#!/bin/bash\n')
        lines.append('ip=$1\n')
        for port_to_test in default_ports:
            lines.append(
                "echo 'Testing connectivity to port: {0} on private ip {1}'\n".format(str(port_to_test), address))
            lines.append(
                'until bash -c "</dev/tcp/$ip/{0}"; [[ "$?" -eq "0" ]];\n'.format(str(port_to_test)))
            lines.append('   do sleep 5;\n')
            lines.append(
                "echo 'Testing connectivity to port: {0} on private ip {1}'\n".format(str(port_to_test), address))
            lines.append('done;\n')
            lines.append("echo 'tested port {0}'\n".format(str(port_to_test)))
        self._file_system.write_lines_to_file(path=script_file_path, lines=lines, chmod=0o777)
        return ['{script_file_path} {ip_address}'.format(
            script_file_path=script_file_path,
            ip_address=address)]


import json
import threading
from logging import Logger
from typing import List

from sidecar.app_instance_identifier import AppInstanceIdentifier
from sidecar.app_instance_service import IAppInstanceService
from sidecar.model.objects import ISidecarConfiguration
from sidecar.services.deployment_outputs_converter import DeploymentOutputsConverter
from sidecar.services.service_updater import IServiceUpdater


class DeploymentOutputsWriter:

    def __init__(self, logger: Logger,
                 app_instance_service: IAppInstanceService,
                 service_updater: IServiceUpdater,
                 config: ISidecarConfiguration):
        self._service_updater = service_updater
        self._app_instance_service = app_instance_service
        self._config = config
        self._logger = logger
        self._lock = threading.RLock()

    def save_application_outputs(self, app_identifier: AppInstanceIdentifier, outputs: str):
        app_name = app_identifier.name
        app_instance_id = app_identifier.infra_id
        try:
            deployment_output = DeploymentOutputsConverter.convert_from_configuration_script(outputs)
        except Exception as exc:
            self._logger.exception(f"application '{app_instance_id}/{app_name}' "
                                   f"deployment output is not valid:\n{outputs}")
            raise exc

        declared_outputs = self._get_application_declared_outputs(app_name)
        self._filter_redundant_deployment_outputs(deployment_output, declared_outputs)

        with self._lock:
            self._app_instance_service.update_deployment_outputs(app_identifier,
                                                                 deployment_output)

    def save_service_outputs(self, service_name: str, output_json: {}):
        try:
            deployment_output = DeploymentOutputsConverter.convert_from_terraform_outputs(output_json)
        except Exception:
            output_str = json.dumps(output_json)
            err = f"service '{service_name}' deployment output is not valid:\n{output_str}"
            self._logger.exception(err)
            raise Exception(err)

        declared_outputs = self._get_service_declared_outputs(service_name)
        self._filter_redundant_deployment_outputs(deployment_output, declared_outputs)
        with self._lock:
            self._service_updater.update_deployment_outputs(service_name, deployment_output)

    @staticmethod
    def _filter_redundant_deployment_outputs(deployment_output: dict, declared_outputs: List[str]):
        deployment_output_names = list(deployment_output.keys())
        for deployment_output_name in deployment_output_names:
            if deployment_output_name not in declared_outputs:
                deployment_output.pop(deployment_output_name, None)

    def _get_service_declared_outputs(self, service_name: str) -> List[str]:
        outputs = next((s.outputs for s in self._config.services if s.name == service_name))
        if outputs:
            return outputs
        return []

    def _get_application_declared_outputs(self, application_name: str) -> List[str]:
        outputs = next((a.outputs for a in self._config.apps if a.name == application_name))
        if outputs:
            return outputs
        return []



import importlib
import pkgutil
from inspect import getmembers, isclass, isabstract
from logging import Logger

from sidecar.aws_session import AwsSession
from sidecar.cloud_logger.file_logger import ICloudLogger
from sidecar.model.objects import ISidecarConfiguration
from .. import cloud_logger
from . import FileLogger


class CloudLoggerFactory(object):
    def __init__(self, config: ISidecarConfiguration, logger: Logger, logger_session: AwsSession, file_logger: FileLogger):
        self.config = config
        self.logger = logger
        self.cloud_logger = None
        self.logger_session = logger_session
        self.file_logger = file_logger

        for path, pkg_name, is_package in pkgutil.iter_modules(cloud_logger.__path__):
            if is_package and pkg_name == config.provider:
                pkg = importlib.import_module(".{}".format(pkg_name), package=cloud_logger.__name__)
                classes = getmembers(pkg, lambda m: isclass(m) and not isabstract(m) and issubclass(m, ICloudLogger))
                if len(classes) > 0:
                    self.cloud_logger = classes[0][1]
                break

    def create_instance(self):
        if self.cloud_logger is not None:
            return self.cloud_logger(self.config, self.logger, self.logger_session, self.file_logger)
        else:
            return self.file_logger



import json
from logging import Logger

from sidecar.app_instance_identifier import AppInstanceIdentifier
from sidecar.app_instance_identifier_creator import IAppInstanceIdentifierCreator
from sidecar.kub_api_pod_reader import KubApiPodReader
from sidecar.kub_api_service import IKubApiService
from sidecar.utils import Utils


class KubAppInstanceIdentifierCreator(IAppInstanceIdentifierCreator):
    def __init__(self, kub_api_service: IKubApiService, logger: Logger) -> None:
        super().__init__(logger=logger)
        self._kub_api_service = kub_api_service

    def create(self, app_name: str, instance_id: str) -> AppInstanceIdentifier:

        container_id = Utils.retry_on_exception(
            func=lambda: self._try_get_container_id(app_name=app_name, instance_id=instance_id),
            logger=self._logger,
            logger_msg=f"getting container id for app '{app_name}' in pod '{instance_id}'",
            timeout_in_sec=10,
            interval_in_sec=1)
        return AppInstanceIdentifier(name=app_name, infra_id=container_id)

    def _try_get_container_id(self, app_name: str, instance_id: str) -> str:
        # using pod name as instance id and not uid because it can easily be sent from the pod ($HOSTNAME) and
        # because it is more efficient to query by.
        # the advantage of uid is that it is unique over time whereas the pod name isn't, but it should be ok since we
        # only use the pod name to find the pod momentarily and then proceed to identify the instance by container id
        pod_json = self._kub_api_service.try_get_pod_json_by_pod_name(pod_name=instance_id)
        if not pod_json:
            raise Exception("pod with name '{NAME}' not found".format(NAME=instance_id))

        # container id is enough as infra identifier because it is supposed to be unique (across space and time),
        # meaning that even if there are multiple instances of the same app on different pods -
        # they should have different container ids
        container_id = KubApiPodReader.safely_get_container_id_for_app(app_name=app_name, pod=pod_json)
        if not container_id:
            raise Exception("container id for app '{APP}' not found on pod '{NAME}':\nPod Json: {POD_JSON}"
                            .format(APP=app_name,
                                    NAME=KubApiPodReader.get_pod_name(pod_json),
                                    POD_JSON=json.dumps(pod_json, indent=4)))
        return container_id



from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.cosmosdb import CosmosDB
from azure.mgmt.network import NetworkManagementClient
from azure.mgmt.resource import ResourceManagementClient
from msrest.authentication import BasicTokenAuthentication
from msrestazure.azure_active_directory import MSIAuthentication, ServicePrincipalCredentials
from azure.mgmt.storage import StorageManagementClient
from azure.cosmosdb.table import TableService


class AzureCredentialsProvider:

    def __init__(self, management_resource_group: str, subscription_id: str, application_id: str,
                 application_secret: str, tenant_id: str, managed_identity_client_id):
        self.management_resource_group = management_resource_group
        self.subscription_id = subscription_id
        self.application_id = application_id
        self.application_key = application_secret
        self.tenant = tenant_id
        self._managed_identity_client_id = managed_identity_client_id

    def get_credentials(self) -> BasicTokenAuthentication:
        # fallback for backwards compatibility
        if not self._managed_identity_client_id:
            return ServicePrincipalCredentials(client_id=self.application_id,
                                               secret=self.application_key,
                                               tenant=self.tenant)
        return MSIAuthentication(client_id=self._managed_identity_client_id)


class AzureClientsManager:

    def __init__(self, credentials_provider: AzureCredentialsProvider):
        self._credentials_provider = credentials_provider
        self._credentials = self._credentials_provider.get_credentials()
        self._cosmos_client = None
        self._compute_client = None
        self._table_service = None
        self._storage_client = None
        self._network_client = None
        self._resource_client = None

    @property
    def management_resource_group(self):
        return self._credentials_provider.management_resource_group

    @property
    def resource_client(self):
        if self._resource_client is None:
            self._resource_client = ResourceManagementClient(self._credentials, self._credentials_provider.subscription_id)
        return self._resource_client

    @property
    def network_client(self):
        if self._network_client is None:
            self._network_client = NetworkManagementClient(self._credentials, self._credentials_provider.subscription_id)
        return self._network_client

    @property
    def cosmos_db_client(self):
        if self._cosmos_client is None:
            self._cosmos_client = CosmosDB(self._credentials, self._credentials_provider.subscription_id)
        return self._cosmos_client

    @property
    def compute_client(self):
        if self._compute_client is None:
            self._compute_client = ComputeManagementClient(self._credentials,
                                                           self._credentials_provider.subscription_id)
        return self._compute_client

    @property
    def storage_client(self):
        if self._storage_client is None:
            self._storage_client = StorageManagementClient(self._credentials,
                                                           self._credentials_provider.subscription_id)
        return self._storage_client

    @property
    def table_service(self):
        if self._table_service is None:
            self._table_service = TableService(account_name=self.management_resource_group.replace("-", ""),
                                               account_key=self._get_storage_key())

        return self._table_service

    def _get_storage_key(self):
        storage_keys = self.storage_client.storage_accounts.list_keys(self.management_resource_group,
                                                                      self.management_resource_group.replace("-", ""))
        storage_keys = {v.key_name: v.value for v in storage_keys.keys}
        return storage_keys['key1']
